/*  *********************************************************************** */
/*  *********************************************************************** */

/*  FILE set.h - *SAT 1.2 */
/*  Memoizing with sets of sets */

/*  *********************************************************************** */
/*  *********************************************************************** */

#include "util.h"
#include "var_set.h"

#include "dp_sat.h"

#ifndef MEMO
#define MEMO

/*  #define TRACE_ DEP */

/* Constants for the data structure 
*/
#define MAX_FLAGS   5      /* No more than 5 flags */
#define MAX_STATS  10      /* No more than 10 statistics */
#define MAX_DEEP  100      /* Maximum depth for "level" caching */

/* Flags */
#define COMPACT_STORAGE    0 /* Try to factorize storage */
#define TRASH_SUBSETS      1 /* Try to trash subsets only */
#define GET_DEPENDENCIES   2 /* Store dependencies and use them to reduce alphas and avoid failed tests */
#define LEVEL_CACHE        3 /* Cache every modal "level" independently */

/* Statistics */
#define MEMO_ACCESS       0  /* # of searches for satisfiable subtests in cache */
#define MEMO_SUCCESS      1  /* # of satisfiable subtests found in cache */
#define DEP_ACCESS        2  /* # of searches for dependencies to reduce alphas */
#define DEP_SUCCESS_PRUNE 3  /* # of alphas actually pruned by dependency search */
#define DEP_SUCCESS_UNSAT 4  /* # of unsatisfiable subtests found in cache */
#define MEMORY            5  /* Bytes allocated */
#define MEMO_STORAGE      6  /* Number of sets stored */
#define MEMO_TRASHING     7  /* Number of sets trashed */
#define DEP_STORAGE       8  /* Number of dependencies stored */
#define DEP_TRASHING      9  /* Number of dependencies trashed */

/* Expected ratio sets vs. dependencies */
#define DEP_FACTOR    1

/* The memoizing data structure
 */

/* A single modal "level" */
typedef struct int_memo {
  var_set_t **memo_array;        /* the data sets */
  int       cur_set;             /* index to the next data set */
  int       growth;              /* 0 = fast growth, 1 = slow growth */  
  int       compact;             /* 0 = add, 1 = stay compact */
  
  var_set_t **dep_array;         /* the dependency sets */
  int       *dep_beta;           /* the dependency labels */
  int       cur_dep;             /* index to the next dependency set */

  int       trashing_sets;       /* Sets and dependencies trashing flag */ 
  int       trashing_deps;

} int_memo_t;

typedef struct memo {
  int_memo_t *memo_of_level[MAX_DEEP];  /* caches in every modal "level" */

  int        level;                     /* current modal "level" */
  int        dim;                       /* initial dimensions of every cache */ 
  int        flags[MAX_FLAGS];          /* flags */
  int        stats[MAX_STATS];          /* global statistics */     
} memo_t;

/* *********************************************************************** */
/* Initialize and clear memoizing */
/* *********************************************************************** */
void Init_memoS(int dim);
void Clear_memoS();
void Force_no_memoS();

/* *********************************************************************** */
/* Accessing the structure */
/* *********************************************************************** */
/* Test if the structure is being used; return 1 if so, 0 otherwise */
int Use_memoS();

/* Set & get flags */
void Setf_memoS(int flag);
void Clearf_memoS(int flag);
int  Getf_memoS(int flag);

/* Get stats */
int Stats_memoS(int stat);
int Get_level_memoS();

/* Increment and decrement modal level */
void Inc_level_memoS();
void Dec_level_memoS();

/* Returns a sublist of beta_lsList_ptr containing betas that are not 
   tested with the current set (or any superset) of the alphas yet. 
   Returns NIL(lsList) when all of the betas were already tested!
*/
lsList *Test_memoS(lsList *alpha_lsList_ptr, lsList *beta_lsList_ptr);

/* Stores a list of alphas that were tested with the 
   corresponding beta. 
   If COMPACT_STORAGE is set, the routine assumes that the beta should be 
   added in the current set. alphas ARE NOT CHECKED!!!
   If TRASH_SUBSETS is set, the routine alternates a "fast" (destructive)
   growth mode and a "slow" (conservative) growth mode. */
void Store_memoS(lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr);

/* Returns a sublist of alpha_lsList_ptr containing dependency-free alphas. 
   If at least one of the betas was tested inconsistent with the 
   current set (or any subset) of the alphas the return value is NIL(lsList).
   In this case rbeta_vertex_ptr_ptr may be used to collect the beta that
   gave the inconsistency */
lsList *Test_depS(lsList *alpha_lsList_ptr, lsList *beta_lsList_ptr,
		  vertex_t **rbeta_vertex_ptr_ptr);

/* Remembers that the list of alphas implies the beta, i.e. it
   stores a dependency (always fast growth) */
void Store_depS(lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr); 

#endif






